/*
 @author: Sascha Mertens
 based on an "XML to JTree" example given by
 Thomas Darimont on
 https://www.tutorials.de/threads/xml-baum-in-jtree-abbilden.171731/

 */
package xpathtester;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;

import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.JTextComponent;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import javax.swing.tree.TreePath;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.xpath.*;

public class JXPathTester extends JFrame {

    private JTree jtree;
    private JTextField tfPath = new JTextField(80);
    private JLabel lblResult = new JLabel("No XPath entered");
    private MyTreeNode jtreeRoot;
    private Document xmlDocument = null;
    private Document xmlDocumentNoWhitespace = null;
    private Document xmlDocumentWhitespace = null;

    private XPath xpath = XPathFactory.newInstance().newXPath();
    private Color background = new Color(235, 240, 240);
    private boolean ignoreWhitespaceNodes = false;
    private JCheckBoxMenuItem miWhitespace;

    public JXPathTester() {
        initComponents();
        try {
            jtreeRoot = makeTreeFromXMLStream(
                    getClass().getResourceAsStream("/example/messung.xml"));
            jtree.setModel(new DefaultTreeModel(jtreeRoot));
        } catch (IOException ex) {
            Logger.getLogger(JXPathTester.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public JXPathTester(String filename) {
        initComponents();
        loadFile(new File(filename));
    }

    private void initComponents() {
        // <editor-fold defaultstate="collapsed">
        int STRUT = 8;
        setTitle("JXPathTester");

        setDefaultCloseOperation(EXIT_ON_CLOSE);
        BorderLayout bl = new BorderLayout(20, 10);
        setLayout(bl);
        // Menu
        JMenuBar mb = new JMenuBar();
        JMenu m1 = new JMenu("File");
        JMenu m2 = new JMenu("Help");
        JMenu m3 = new JMenu("Options");

        JMenuItem mi = new JMenuItem("Open File");
        JMenuItem miabout = new JMenuItem("About");
        miWhitespace
                = new JCheckBoxMenuItem("Ignore Whitespace Nodes", ignoreWhitespaceNodes);

        miabout.addActionListener(alAbout);
        mi.addActionListener(alFileLoad);
        miWhitespace.addActionListener(alIgnoreWhitespaces);
        m1.add(mi);
        m2.add(miabout);
        m3.add(miWhitespace);
        mb.add(m1);
        mb.add(m3);
        mb.add(m2);
        add(mb, BorderLayout.NORTH);
        JPanel panMain = new JPanel(new BorderLayout());
        add(panMain, BorderLayout.CENTER);
        panMain.setBorder(new EmptyBorder(STRUT, STRUT, STRUT, STRUT));
        //-------------------------------------------------------
        JPanel pan1 = new JPanel(new BorderLayout(0, 10));
        pan1.setBorder(new EmptyBorder(0, 0, 2 * STRUT, 0));
        JLabel l1 = new JLabel("Enter Path");
        pan1.add(l1, BorderLayout.NORTH);
        pan1.add(tfPath, BorderLayout.CENTER);
        tfPath.addKeyListener(alKeyUp);
        pan1.add(lblResult, BorderLayout.SOUTH);
        panMain.add(pan1, BorderLayout.NORTH);
        // --------- Tree -----   
        jtree = new JTree();
        jtree.setEditable(false);

        JScrollPane sc = new JScrollPane(jtree);
        jtree.setCellRenderer(new MyTreeRenderer());
        sc.setBackground(background);
        jtree.setBackground(background);
        panMain.add(sc, BorderLayout.CENTER);
//        
        int h = Toolkit.getDefaultToolkit().getScreenSize().height;
        setMinimumSize(new Dimension(800, h * 2 / 3));
        pack();

        setVisible(true);
        setMinimumSize(getSize());
//        pan1.setMaximumSize(pan1.getSize());
        setLocation(h / 4, h / 8);
// </editor-fold>
    }

//-----------------------------------------------------------------------------
//--------------------  Tree and XPath ---------------------------
//-----------------------------------------------------------------------------
    private void xpathAnwenden() {
        String pfad = tfPath.getText();
        int results = 0;
        MyTreeNode.unmarkAll();
        if (!pfad.isEmpty()) {
            if (pfad.equals("/")) {
                System.out.println(jtreeRoot);
                jtreeRoot.setMarked(true);
                results++;
            } else {
                try {
                    XPathExpression expr = xpath.compile(pfad);
                    NodeList nodeList = (NodeList) expr.evaluate(xmlDocument, XPathConstants.NODESET);
                    for (int i = 0; i < nodeList.getLength(); ++i) {
                        Node node = nodeList.item(i);
                        results++;
                        makeResultVisible(node);
                    } // for

                } catch (XPathExpressionException ex) {
                    lblResult.setText("Input not valid/complete");
                }
            }
            lblResult.setText("Your path selects " + results
                    + " Element" + (results > 0 ? "s" : ""));
        }
        jtree.repaint();
    } // pathAnwenden

    private void makeResultVisible(Node node) {
        MyTreeNode treeNode = MyTreeNode.getTreeNodeCorrespondingTo(node);
        if (treeNode != null) {
            treeNode.setMarked(true);
            expandSelectedNode(node, treeNode);
        } else {
            System.out.printf("Something looks wrong,node  %s "
                    + "should have a corresponding treenode", node);
        } // if
    } // makeRsultVisible

    private void expandSelectedNode(Node node, DefaultMutableTreeNode treeNode) {
        if (node.getNodeType() != Node.ELEMENT_NODE) {
            //System.out.println(node.getParentNode());
            if (treeNode.getParent() instanceof MyTreeNode) {
                jtree.expandPath(new TreePath(
                        ((MyTreeNode) treeNode.getParent()).getPath()));
            } else { // shouldn't happen...
                jtree.expandPath(new TreePath(
                        treeNode.getPreviousNode().getPath()));
            }
        } else {
            jtree.expandPath(new TreePath(treeNode.getPath()));
        }
    } // expand

    private MyTreeNode makeTreeFromXMLStream(InputStream fstream) throws IOException {
        try {
            DocumentBuilderFactory factory
                    = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            factory.setNamespaceAware(false);
            factory.setIgnoringComments(false);

            xmlDocument = builder.parse(fstream);
            xmlDocumentWhitespace = xmlDocument;
            xmlDocumentNoWhitespace = copyDocument(xmlDocument);

            removeWhitespace(xmlDocumentNoWhitespace.getDocumentElement(), xmlDocumentNoWhitespace);

        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();

        }
        return buildJtree(xmlDocument);
    } // makeTree

    public Document copyDocument(Document source) {
        try {
            TransformerFactory tfactory = TransformerFactory.newInstance();
            Transformer tx = tfactory.newTransformer();
            DOMResult result = new DOMResult();

            tx.transform(new DOMSource(source), result);
            return (Document) result.getNode();
        } catch (TransformerConfigurationException ex) {
            Logger.getLogger(JXPathTester.class.getName()).log(Level.SEVERE, null, ex);
        } catch (TransformerException ex) {
            Logger.getLogger(JXPathTester.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public void removeWhitespace(Node n, Node parent) {
        //System.out.println("." + n);
        if (n.hasChildNodes()) {
            NodeList list = n.getChildNodes();
            for (int i = list.getLength() - 1; i >= 0; i--) {
                Node child = list.item(i);
                removeWhitespace(child, n);
            } // for
        } else {
            if (n.getNodeType() == Node.TEXT_NODE) {
                if (n.getNodeValue().trim().isEmpty()) {
                    System.out.println("remove " + n);
                    parent.removeChild(n);
                };
            }
        } // else
    } // remove

    public MyTreeNode buildJtree(Document doc) {
        // Now build a jtree
        MyTreeNode.clearList();
        MyTreeNode treeRoot = new MyTreeNode(null, "/");
        /*Node start = getRootNode(xmlDocument);
         treeWalk(start, 0, treeRoot);*/
        treeWalk(doc.getDocumentElement(), 0, treeRoot);
        return treeRoot;
    } // getTreeFromXML

    public void treeWalk(
            Node node,
            int level,
            MyTreeNode jtreeParentNode) {
        if (node.getNodeType() == Node.COMMENT_NODE) {
            return;
        }
        if (node.hasChildNodes() || node.hasAttributes()) {
            MyTreeNode newTreeNode
                    = new MyTreeNode(node, node.getNodeName());
            jtreeParentNode.add(newTreeNode);
            attribute(node, newTreeNode);
            NodeList list = node.getChildNodes();
            int len = list.getLength();
            for (int i = 0; i < len; i++) {
                treeWalk(list.item(i), level + 1, newTreeNode);
            } // for

        } else {
            String nv = node.getNodeValue();
            if (node.getNodeValue() != null) {
                String text = "#text:" + node.getNodeValue();
                jtreeParentNode.add(new MyTreeNode(node, text));
            } // textknoten nicht null
        } // else

    } // walktree

    private void attribute(Node node, MyTreeNode treenode) {
        if (node.hasAttributes()) {
            NamedNodeMap attMap = node.getAttributes();
            for (int i = 0; i < attMap.getLength(); i++) {
                Node attnode = attMap.item(i);
                treenode.add(new MyTreeNode(attnode, "@"
                        + attnode.getNodeName() + "=" + attnode.getNodeValue()));
            }
        }
    } // attribute

//-----------------------------------------------------------------------------
//------------------------ File Operations ----------------------   
//-----------------------------------------------------------------------------
    private String getXmlFromFile(String file) {
        Scanner s;
        try {
            s = new Scanner(new File(file));
            StringBuilder sb = new StringBuilder();
            while (s.hasNext()) {
                sb.append(s.nextLine());
            }
            return sb.toString().replace(System.getProperty("line.separator"), "").replaceAll("> +<", "><");

        } catch (FileNotFoundException ex) {
            Logger.getLogger(JXPathTester.class
                    .getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    private void loadFile() {
        File f = new File(System.getProperty("user.dir"));

        JFileChooser fc = new JFileChooser(f);
        fc.setFileFilter(new FileNameExtensionFilter("XML-File", "xml"));
        int opt = fc.showOpenDialog(getContentPane());
        if (opt == JFileChooser.APPROVE_OPTION) {
            loadFile(fc.getSelectedFile());
        }
    } // loadfile

    private void loadFile(File file) {
        jtreeRoot = getTreeFromFile(file);
        jtree.setModel(new DefaultTreeModel(jtreeRoot));
    }

    private MyTreeNode getTreeFromFile(File file) {
        try {
            return makeTreeFromXMLStream(new FileInputStream(file));
        } catch (IOException ex) {
            fehlermeldung("Fehler beim Laden der Datei \n"
                    + file.getAbsolutePath() + "\n\n" + ex.getMessage());
            ex.printStackTrace();

        }
        return new MyTreeNode(null, "EMPTY");
    }

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
    private void fehlermeldung(String text) {
        JOptionPane.showMessageDialog(rootPane, text, "Error", JOptionPane.ERROR_MESSAGE);
    }
    // ActionListeners
    private ActionListener alFileLoad = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            loadFile();
        }
    };
    private ActionListener alAbout = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            JOptionPane.showMessageDialog(rootPane,
                    "JXPathTester\nVersion 1.0 (Nov. 2014)\nby S.Mertens", "About",
                    JOptionPane.INFORMATION_MESSAGE);
        }
    };
    private ActionListener alIgnoreWhitespaces = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            ignoreWhitespaceNodes = miWhitespace.isSelected();
            if (ignoreWhitespaceNodes) {
                xmlDocument = xmlDocumentNoWhitespace;
            } else {
                xmlDocument = xmlDocumentWhitespace;
            }
            jtreeRoot = buildJtree(xmlDocument);
            jtree.setModel(new DefaultTreeModel(jtreeRoot));
            xpathAnwenden();
        }
    };
    private KeyAdapter alKeyUp = new KeyAdapter() {
        @Override
        public void keyReleased(KeyEvent e) {
            xpathAnwenden();
        }
    }; // alKEyup

// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------
    public static void main(String[] args) {
        if (args.length > 0) {
            new JXPathTester(args[0]);
        } else {
            new JXPathTester();

        }
    } // main

} // class
